import os.path
from typing import Dict, Any, Optional

import pandas as pd
from torch import nn
from wandb.wandb_torch import torch

from torch_mist.utils.logging.logger.base import Logger


class PandasLogger(Logger):
    def __init__(
        self,
        log_dir: str = ".",
        log_name: Optional[str] = None,
        log_every: int = 10,
    ):
        super().__init__(log_dir=log_dir, log_every=log_every)
        self.__log = []
        if log_name is None:
            log_name = "log.csv"
        if not log_name.endswith(".csv"):
            log_name = log_name + ".csv"
        self.log_name = log_name

    def _log(
        self, data: Any, name: str, iteration: int, epoch: int, split: str
    ):
        if not isinstance(data, Dict):
            data = {
                "mean": data,
            }
        data["name"] = name
        data["split"] = split
        data["iteration"] = iteration
        data["epoch"] = epoch
        self.__log.append(data)

    def get_log(self) -> pd.DataFrame:
        return pd.DataFrame(self.__log)

    def _reset_log(self):
        self.__log = []

    def save_log(self):
        log = self.get_log()
        log_path = os.path.join(self.log_dir, self.log_name)
        log.to_csv(log_path)



import inspect
from typing import Optional, Any, Union, Tuple, Dict

import torch
from torch.utils.data import DataLoader, random_split

from torch_mist.estimators import MIEstimator, TransformedMIEstimator
from torch_mist.estimators.hybrid import PQHybridMIEstimator
from torch_mist.utils.batch import unfold_samples, move_to_device
from torch_mist.utils.data import SampleDataset, SameAttributeDataLoader
from torch_mist.utils.data.loader import sample_same_value


def make_default_dataloaders(
    x: Optional[torch.Tensor] = None,
    y: Optional[torch.Tensor] = None,
    train_loader: Optional[Any] = None,
    valid_loader: Optional[Any] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 8,
) -> Tuple[DataLoader, DataLoader]:
    if (x is None) != (y is None):
        raise ValueError(
            "Either both x and y need to be specified or neither."
        )

    if not ((x is None) ^ (train_loader is None)):
        raise ValueError(
            "Either both x and y or the train_loader need to be specified."
        )
    if not (x is None):
        if batch_size is None:
            raise ValueError("Please specify a value for batch_size.")

        # Make the dataloaders from the samples
        train_set = SampleDataset({"x": x, "y": y})
        if valid_percentage > 0:
            if not (valid_loader is None):
                raise ValueError(
                    "The valid_loader can't be specified when x and y are used"
                )

            n_valid = int(len(train_set) * valid_percentage)
            train_set, val_set = random_split(
                train_set, [len(train_set) - n_valid, n_valid]
            )
            valid_loader = DataLoader(
                val_set,
                batch_size=batch_size,
                num_workers=num_workers,
            )

        train_loader = DataLoader(
            train_set,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
        )

    return train_loader, valid_loader


def make_dataloaders(
    estimator: MIEstimator,
    x: Optional[torch.Tensor] = None,
    y: Optional[torch.Tensor] = None,
    train_loader: Optional[Any] = None,
    valid_loader: Optional[Any] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 8,
) -> Tuple[DataLoader, Optional[DataLoader]]:
    train_loader, valid_loader = make_default_dataloaders(
        x=x,
        y=y,
        train_loader=train_loader,
        valid_loader=valid_loader,
        valid_percentage=valid_percentage,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # Check if we need to modify the dataloader to sample the same attributes
    _estimator = estimator
    transforms = []
    while isinstance(_estimator, TransformedMIEstimator):
        transforms.append(_estimator.transforms)
        _estimator = estimator.base_estimator

    # If required, change the data-loader to sample batches with the same attribute only
    if isinstance(_estimator, PQHybridMIEstimator):
        neg_samples = _estimator.neg_samples
        device = next(iter(estimator.parameters())).device

        def compute_attributes(samples):
            variables = unfold_samples(samples)
            variables = move_to_device(variables, device)
            assert "y" in variables
            y = variables["y"]
            for transform in transforms:
                y = transform["y->y"](y)
            return _estimator.generative_estimator.transforms["y->y"](
                y
            ).data.cpu()

        if not isinstance(train_loader, SameAttributeDataLoader):
            train_loader = sample_same_value(
                train_loader, compute_attributes, neg_samples=neg_samples
            )

        if not isinstance(valid_loader, SameAttributeDataLoader) and not (
            valid_loader is None
        ):
            valid_loader = sample_same_value(
                valid_loader, compute_attributes, neg_samples=neg_samples
            )

    return train_loader, valid_loader



from typing import Type, Optional, Dict, Any, Union, Tuple, List, Callable

import numpy as np
import torch
from torch.optim import Optimizer
from tqdm.autonotebook import tqdm
from torch.optim import Adam
from torch.optim.lr_scheduler import (
    SequentialLR,
    LinearLR,
    CosineAnnealingLR,
    ConstantLR,
    LRScheduler,
)
from torch.utils.data import DataLoader

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.batch import unfold_samples, move_to_device

from torch_mist.utils.evaluation import evaluate_mi
from torch_mist.utils.logging import PandasLogger
from torch_mist.utils.logging.logger.base import Logger, DummyLogger
from torch_mist.utils.misc import make_dataloaders
from torch_mist.utils.train.utils import RunTerminationManager


def _instantiate_optimizer(
    estimator: MIEstimator,
    max_epochs: int,
    iterations_per_epoch: int,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
) -> Tuple[Optimizer, LRScheduler]:
    params = [
        {"params": params}
        for params in estimator.parameters()
        if params.requires_grad
    ]

    if optimizer_params is None:
        optimizer_params = {"lr": 5e-4}

    opt = optimizer_class(params, **optimizer_params)

    # Cosine annealing with initial linear warmup
    if lr_annealing:
        max_iterations = iterations_per_epoch * max_epochs
        if not 0 <= warmup_percentage <= 1:
            raise ValueError("Warmup percentage must be between 0 and 1")
        warmup_iterations = int(
            iterations_per_epoch * max_epochs * warmup_percentage
        )
        lr_scheduler = SequentialLR(
            opt,
            [
                LinearLR(
                    opt, start_factor=1e-2, total_iters=warmup_iterations
                ),
                CosineAnnealingLR(
                    opt,
                    T_max=max_iterations - warmup_iterations,
                    eta_min=1e-5,
                ),
            ],
            milestones=[warmup_iterations],
        )
    else:
        lr_scheduler = ConstantLR(opt, 1.0)

    return opt, lr_scheduler


def train_epoch(
    estimator: MIEstimator,
    train_loader: DataLoader,
    opt: Optimizer,
    lr_scheduler: LRScheduler,
    device: Union[str, torch.device],
    logger: Optional[Logger] = None,
    tqdm_iteration: Optional[tqdm] = None,
    fast_train: bool = False,
    train_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    max_iterations: Optional[int] = None,
):
    with logger.train():
        estimator.train()
        with logger.epoch():
            for samples in train_loader:
                variables = unfold_samples(samples)
                variables = move_to_device(variables, device)

                if max_iterations:
                    if logger._iteration >= max_iterations:
                        break

                with logger.iteration():
                    with logger.logged_methods(
                        estimator, train_logged_methods
                    ):
                        loss = estimator(**variables)

                    # Compute the ratio only if necessary
                    if not fast_train and not isinstance(logger, DummyLogger):
                        with logger.logged_methods(
                            estimator, eval_logged_methods
                        ):
                            estimator.mutual_information(**variables)

                    opt.zero_grad()
                    loss.backward()
                    opt.step()
                    lr_scheduler.step()

                if tqdm_iteration:
                    tqdm_iteration.update(1)
                    tqdm_iteration.set_postfix_str(f"loss: {loss}")


def validate(
    estimator: MIEstimator,
    valid_loader: Optional[DataLoader],
    device: Union[str, torch.device],
    logger: Logger,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
) -> float:
    if valid_loader is not None:
        with logger.valid():
            with logger.logged_methods(
                estimator,
                eval_logged_methods,
            ):
                valid_mi = evaluate_mi(
                    estimator=estimator,
                    dataloader=valid_loader,
                    device=device,
                )

            if isinstance(valid_mi, dict):
                valid_mi = sum(valid_mi.values())
    else:
        valid_mi = None

    return valid_mi


def train_mi_estimator(
    estimator: MIEstimator,
    x: Optional[torch.Tensor] = None,
    y: Optional[torch.Tensor] = None,
    train_loader: Optional[Any] = None,
    valid_loader: Optional[Any] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 8,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: Optional[int] = None,
    max_iterations: Optional[int] = None,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    early_stopping: bool = False,
    patience: int = 3,
    delta: float = 0.001,
    fast_train: bool = False,
    train_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
) -> Optional[Any]:
    # Create the training and validation dataloaders
    train_loader, valid_loader = make_dataloaders(
        estimator=estimator,
        x=x,
        y=y,
        train_loader=train_loader,
        valid_loader=valid_loader,
        valid_percentage=valid_percentage,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    if max_epochs is None:
        if max_iterations is None:
            raise ValueError(
                "Please specify either max_epochs or max_iterations"
            )
        max_epochs = int(np.ceil(max_iterations / len(train_loader)))

    opt, lr_scheduler = _instantiate_optimizer(
        estimator=estimator,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        lr_annealing=lr_annealing,
        warmup_percentage=warmup_percentage,
        max_epochs=max_epochs,
        iterations_per_epoch=len(train_loader),
    )

    estimator = estimator.to(device)

    default_logger = False
    # If the logger is None, use the default PandasLogger,
    if logger is None:
        logger = PandasLogger()
        default_logger = True
    # If False, instantiate a DummyLogger, which does not store any quantity
    elif logger is False:
        logger = DummyLogger()
        default_logger = True

    # If nothing is specified, log the loss and mutual information
    if train_logged_methods is None:
        train_logged_methods = ["loss"]
    if eval_logged_methods is None:
        eval_logged_methods = ["mutual_information"]

    run_manager = RunTerminationManager(
        early_stopping=early_stopping,
        delta=delta,
        patience=patience,
        verbose=verbose,
        max_iterations=max_iterations,
        maximize=estimator.lower_bound,
        minimize=estimator.upper_bound,
    )

    tqdm_epochs = (
        tqdm(total=max_epochs, desc="Epoch", position=1) if verbose else None
    )
    tqdm_iteration = (
        tqdm(total=len(train_loader), desc="Iteration", position=1)
        if verbose
        else None
    )

    for epoch in range(max_epochs):
        if tqdm_epochs:
            tqdm_iteration.reset()

        train_epoch(
            estimator=estimator,
            train_loader=train_loader,
            opt=opt,
            lr_scheduler=lr_scheduler,
            logger=logger,
            tqdm_iteration=tqdm_iteration,
            device=device,
            fast_train=fast_train,
            train_logged_methods=train_logged_methods,
            eval_logged_methods=eval_logged_methods,
            max_iterations=max_iterations,
        )

        valid_mi = validate(
            estimator=estimator,
            valid_loader=valid_loader,
            device=device,
            logger=logger,
            eval_logged_methods=eval_logged_methods,
        )

        if tqdm_epochs:
            if valid_mi:
                tqdm_epochs.set_postfix_str(f"valid_mi: {valid_mi}")
            tqdm_epochs.update(1)

        if run_manager.should_stop(
            iteration=logger._iteration, valid_mi=valid_mi
        ):
            break

    log = logger.get_log()

    if default_logger:
        logger.clear()

    return log



from typing import Optional, Any, Union, Type, Dict, Tuple, List

import torch
import numpy as np
import pandas as pd
from torch.optim import Optimizer, Adam
from torch.utils.data import DataLoader

from torch_mist.estimators.base import MIEstimator
from torch_mist.estimators.factories import instantiate_estimator
from torch_mist.utils.batch import unfold_samples
from torch_mist.utils.logging import PandasLogger
from torch_mist.utils.logging.logger.base import Logger, DummyLogger
from torch_mist.utils.train.mi_estimator import train_mi_estimator
from torch_mist.utils.evaluation import evaluate_mi


def _infer_dim(
    x: Optional[torch.Tensor],
    y: Optional[torch.Tensor],
    train_loader: Optional[DataLoader],
) -> Tuple[int, int]:
    if x is None:
        batch = next(iter(train_loader))
        variables = unfold_samples(batch)
        if not ("x" in variables):
            raise ValueError(
                "Each batch must consist of a tuple (x,y) or a dictionary containing a key for 'x'"
            )
        x_dim = variables["x"].shape[-1]
        if not ("y" in variables):
            raise ValueError(
                "Each batch must consist of a tuple (x,y) or a dictionary containing a key for 'y'"
            )
        y_dim = variables["y"].shape[-1]
    else:
        x_dim = x.shape[-1]
        y_dim = y.shape[-1]

    return x_dim, y_dim


def estimate_mi(
    estimator_name: str,
    x: Optional[Union[torch.Tensor, np.array]] = None,
    y: Optional[Union[torch.Tensor, np.array]] = None,
    train_loader: Optional[DataLoader] = None,
    valid_loader: Optional[DataLoader] = None,
    test_loader: Optional[DataLoader] = None,
    valid_percentage: float = 0.2,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: Optional[int] = None,
    max_iterations: Optional[int] = None,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
    batch_size: Optional[int] = 128,
    evaluation_batch_size: Optional[int] = None,
    num_workers: int = 8,
    early_stopping: bool = False,
    patience: int = 3,
    delta: float = 0.001,
    return_estimator: bool = False,
    fast_train: bool = False,
    hidden_dims: List[int] = [128, 64],
    **kwargs,
) -> Union[
    float,
    Tuple[float, pd.DataFrame],
    Tuple[float, MIEstimator],
    Tuple[float, MIEstimator, pd.DataFrame],
]:
    if not (x is None):
        if len(x.shape) == 1:
            x = x.reshape(-1, 1)
    if not (y is None):
        if len(y.shape) == 1:
            y = y.reshape(-1, 1)

    x_dim, y_dim = _infer_dim(x=x, y=y, train_loader=train_loader)

    if verbose:
        print(f"Instantiating the {estimator_name} estimator")
    estimator = instantiate_estimator(
        estimator_name=estimator_name,
        x_dim=x_dim,
        y_dim=y_dim,
        hidden_dims=hidden_dims,
        **kwargs,
    )
    if verbose:
        print(estimator)
        print("Training the estimator")

    if evaluation_batch_size is None:
        evaluation_batch_size = batch_size

    if logger is None:
        logger = PandasLogger()
    elif logger is False:
        logger = DummyLogger()

    train_log = train_mi_estimator(
        estimator=estimator,
        x=x,
        y=y,
        train_loader=train_loader,
        valid_loader=valid_loader,
        valid_percentage=valid_percentage,
        device=device,
        max_epochs=max_epochs,
        max_iterations=max_iterations,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        verbose=verbose,
        logger=logger,
        lr_annealing=lr_annealing,
        warmup_percentage=warmup_percentage,
        batch_size=batch_size,
        early_stopping=early_stopping,
        patience=patience,
        delta=delta,
        num_workers=num_workers,
        fast_train=fast_train,
    )

    if verbose:
        print("Evaluating the value of Mutual Information")
    if not (test_loader is None):
        x = None
        y = None
    elif x is None:
        print(
            "Warning: using the train_loader to estimate the value of mutual information. Please specify a test_loader"
        )
        test_loader = train_loader

    with logger.test():
        mi_value = evaluate_mi(
            estimator=estimator,
            x=x,
            y=y,
            dataloader=test_loader,
            batch_size=evaluation_batch_size,
            device=device,
            num_workers=num_workers,
        )

    logger.clear()

    if not (train_log is None) and return_estimator:
        return mi_value, estimator, train_log
    elif not (train_log is None):
        return mi_value, train_log
    elif return_estimator:
        return mi_value, estimator
    else:
        return mi_value



import collections
from typing import Optional, Any, List, Union, Dict

import torch
import numpy as np

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.batch import unfold_samples, move_to_device
from torch_mist.utils.misc import make_dataloaders


def evaluate_mi(
    estimator: MIEstimator,
    x: Optional[torch.Tensor] = None,
    y: Optional[torch.Tensor] = None,
    dataloader: Optional[Any] = None,
    device: torch.device = torch.device("cpu"),
    batch_size: Optional[int] = None,
    num_workers: int = 8,
) -> Union[float, Dict[str, float]]:
    dataloader, _ = make_dataloaders(
        estimator=estimator,
        x=x,
        y=y,
        train_loader=dataloader,
        valid_percentage=0,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    estimator.eval()
    estimator = estimator.to(device)

    mis = collections.defaultdict(list)
    unfold = False
    for samples in dataloader:
        variables = unfold_samples(samples)
        variables = move_to_device(variables, device)

        estimation = estimator.mutual_information(**variables)

        if isinstance(estimation, dict):
            for (x_key, y_key), value in estimation.items():
                mis[f"I({x_key};{y_key})"].append(value.item())
        else:
            mis["I(x;y)"].append(estimation.item())
            unfold = True

    mi = {}
    for name, values in mis.items():
        mi[name] = np.mean(values)

    if unfold:
        return mi["I(x;y)"]
    return mi



from typing import Optional


class RunTerminationManager:
    def __init__(
        self,
        early_stopping: bool,
        delta: float,
        patience: int,
        verbose: bool,
        max_iterations: Optional[int],
        maximize: bool = False,
        minimize: bool = False,
    ):
        if early_stopping:
            if not maximize and not minimize:
                raise ValueError(
                    "early_stopping can be used only when maximizing or minimizing"
                )
        self.early_stopping = early_stopping
        self.maximize = maximize
        self.minimize = minimize
        self.best_value = 0
        self.delta = delta
        self.patience = patience
        self.current_patience = patience
        self.verbose = verbose
        self.max_iterations = max_iterations

    def should_stop(self, iteration: int, valid_mi: Optional[float]) -> bool:
        stop = False
        if self.early_stopping:
            improvement = (
                (valid_mi - self.best_value)
                if self.maximize
                else (self.best_value - valid_mi)
            )
            if improvement >= self.delta:
                # Improvement, update best and reset the patience
                self.best_value = valid_mi
                self.current_patience = self.patience
            else:
                self.patience -= 1

            if self.patience < 0:
                if self.verbose:
                    print("No improvements on validation, stopping.")
                stop = True

        if self.max_iterations:
            if iteration >= self.max_iterations:
                stop = True
        return stop
